import sys
import boto3
import re
import time
from botocore.config import Config
from botocore.exceptions import ClientError
import shutil
import os

# Custom Dependencies
from common_utils.scripts import common_utilities
from common_utils.scripts.common_utilities import CommonLogging
logging = CommonLogging.get_logger()
# ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** *****


class CommonS3Utilities:

    MINIO_S3_PROTOCOL_PREFIX = "s3a://"
    AWS_S3_PROTOCOL_PREFIX = "s3://"

    @common_utilities.singleton
    class AwsS3Session:
        def __init__(self, aws_s3_credentials):
            s3_client_config = Config(region_name=aws_s3_credentials["s3_region"], proxies={
                'http': os.environ.get("https_proxy"), 'https': os.environ.get("https_proxy")
            }, proxies_config={"proxy_ca_bundle": os.environ.get("REQUESTS_CA_BUNDLE")})

            self.aws_s3_client = boto3.client('s3', config=s3_client_config,
                                              aws_access_key_id=aws_s3_credentials["s3_access_key"],
                                              aws_secret_access_key=aws_s3_credentials["s3_secret_access_key"],
                                              verify=os.environ.get("REQUESTS_CA_BUNDLE"))

    @common_utilities.singleton
    class MinioS3Client:
        def __init__(self, minio_s3_credentials):
            self.minio_s3_client = boto3.client(
                's3',
                endpoint_url=minio_s3_credentials["s3_host"],
                aws_access_key_id=minio_s3_credentials["s3_access_key"],
                aws_secret_access_key=minio_s3_credentials["s3_secret_access_key"],
                config=Config(signature_version=minio_s3_credentials["s3_signature_version"]),
                region_name=minio_s3_credentials["s3_region"],
                verify=True
            )

    @common_utilities.singleton
    class MinioS3Resource:
        def __init__(self, minio_s3_credentials):
            self.minio_s3_resource = boto3.resource(
                's3',
                endpoint_url=minio_s3_credentials["s3_host"],
                aws_access_key_id=minio_s3_credentials["s3_access_key"],
                aws_secret_access_key=minio_s3_credentials["s3_secret_access_key"],
                config=Config(signature_version=minio_s3_credentials["s3_signature_version"]),
                region_name=minio_s3_credentials["s3_region"],
                verify=os.getenv('REQUESTS_CA_BUNDLE', True)
            )

    @staticmethod
    def ls_objects_from_s3(s3_client, s3_full_path, s3_protocol_prefix):
        s3_bucket = ""
        s3_path_prefix = ""
        try:
            s3_bucket, s3_path_prefix = CommonS3Utilities.get_s3_bucket_and_prefix_from_path(
                s3_full_path, s3_protocol_prefix)

            paginator = s3_client.get_paginator('list_objects_v2')
            pages = paginator.paginate(Bucket=s3_bucket, Prefix=s3_path_prefix)
            s3_objects_list = []
            for page in pages:
                try:
                    for obj in page['Contents']:
                        s3_objects_list.append("{}{}/".format(s3_protocol_prefix, s3_bucket) + obj['Key'])

                        # if s3_incremental_flag:
                        #
                        #     # Filtering files by modified date
                        #     if obj['LastModified'].date() == (datetime.now().date():
                        #         s3_objects_list.append("{}{}/".format(s3_protocol_prefix, s3_bucket) + obj['Key'])
                        #
                        #     # Filtering files by file name which contains prior date
                        #     # prior_date = str((datetime.now() - timedelta(days=1)).date())
                        #     # if obj['Key'].__contains__(prior_date):
                        #     #     s3_objects_list.append("{}{}/".format(s3_protocol_prefix, s3_bucket) + obj['Key'])
                        #
                        # else:
                        #     s3_objects_list.append("{}{}/".format(s3_protocol_prefix, s3_bucket) + obj['Key'])
                except KeyError as e:
                    logging.warning("No files are available in the path, S3_path:%s" % s3_full_path)

            s3_files_list = [s3_object for s3_object in s3_objects_list if s3_object[-1] != '/']

            logging.info("len(s3_objects_list): %s" % (str(len(s3_objects_list))))

            logging.info("len(s3_files_list): %s" % (str(len(s3_objects_list))))

            return s3_files_list

        except Exception as e:
            raise Exception("Exception occurred while listing files from S3 path, Error:%s ,s3_bucket:%s, "
                            "s3_path_prefix:%s" % (str(e), s3_bucket, s3_path_prefix))

    # ##########################################################################################
    @staticmethod
    def create_pre_signed_url(s3_client, s3_bucket, s3_prefix, expiration=180):

        try:
            pre_signed_url = s3_client.generate_presigned_url('get_object',
                                                              Params={'Bucket': s3_bucket,
                                                                          'Key': s3_prefix
                                                                      },
                                                              ExpiresIn=expiration
                                                              )
        except ClientError as e:
            logging.error(e)
            raise Exception(e)

        # The response contains the pre_signed URL
        return pre_signed_url

    @staticmethod
    def read_file_from_s3(s3_resource, s3_full_path, s3_protocol_prefix):

        try:
            s3_bucket, s3_path_prefix = CommonS3Utilities.get_s3_bucket_and_prefix_from_path(s3_full_path,
                                                                                             s3_protocol_prefix)

            print(f"{s3_bucket}, {s3_path_prefix}")
            response = s3_resource.Object(s3_bucket, s3_path_prefix)
            data = response.get()['Body'].read().decode('utf-8')

            return data
        except Exception as e:
            raise Exception("Exception occurred while reading Config File, Error:%s, s3_full_path: %s"
                             % (str(e), s3_full_path))

    @staticmethod
    def s3_write_file(s3_resource, data_bytes, s3_full_path, s3_protocol_prefix):

        s3_bucket, s3_path_prefix = CommonS3Utilities.get_s3_bucket_and_prefix_from_path(s3_full_path,
                                                                                         s3_protocol_prefix)

        logging.info(">> >> >> s3_bucket:%s, s3_prefix:%s" % (s3_bucket, s3_path_prefix))

        try:
            s3_object = s3_resource.Object(s3_bucket, s3_path_prefix)

            s3_object.put(Body=data_bytes)
        except Exception as e:
            raise Exception("Exception occurred while writing File, Error:%s ,s3_bucket:%s, "
                            "s3_path_prefix:%s" % (str(e), s3_bucket, s3_path_prefix))

    # **********Downloading file with Boto3 download********************************
    @staticmethod
    def s3_download_file(s3_resource, s3_full_path, local_temp_dir_base_path, s3_protocol_prefix):

        try:
            s3_bucket, s3_path_prefix = CommonS3Utilities.get_s3_bucket_and_prefix_from_path(s3_full_path,
                                                                                             s3_protocol_prefix)
            s3_bucket_obj = s3_resource.Bucket(s3_bucket)

            path, filename = os.path.split(s3_path_prefix)
            local_temp_dir_path = os.path.join(local_temp_dir_base_path, path)

            if not os.path.exists(local_temp_dir_path):
                os.makedirs(local_temp_dir_path)
            else:
                shutil.rmtree(local_temp_dir_path)
                os.makedirs(local_temp_dir_path)

            local_temp_file_path = os.path.join(local_temp_dir_path, filename)
            s3_bucket_obj.download_file(s3_path_prefix, local_temp_file_path)
            return local_temp_file_path

        except Exception as e:

            raise Exception("Exception occurred while downloading files from S3 path, Error:%s, s3_full_path:%s, "
                            "local_temp_dir_base_path:%s" % (str(e), s3_full_path, local_temp_dir_base_path))

    @staticmethod
    def s3_upload_file(s3_client, local_temp_file_path,
                       local_temp_dir_base_path, s3_base_path, s3_protocol_prefix):
        try:
            s3_path_prefix = os.path.relpath(local_temp_file_path, local_temp_dir_base_path)
            s3_full_path = os.path.join(s3_base_path, s3_path_prefix)

            s3_bucket, s3_path_prefix = CommonS3Utilities.get_s3_bucket_and_prefix_from_path(
                s3_full_path, s3_protocol_prefix)

            s3_client.upload_file(local_temp_file_path, s3_bucket, s3_path_prefix)

        except Exception as e:
            raise Exception("Exception occurred while uploading file from local path, Error:%s, "
                            "local_temp_file_path:%s, s3_base_path:%s" % (str(e), local_temp_file_path, s3_base_path))

    @staticmethod
    def s3_move_file(s3_source_file_full_path, s3_target_file_full_path, s3_resource,  s3_protocol_prefix):

        try:
            s3_source_bucket, s3_source_file_path_prefix = CommonS3Utilities.get_s3_bucket_and_prefix_from_path(
                s3_source_file_full_path, s3_protocol_prefix)

            s3_target_bucket, s3_target_file_path_prefix = CommonS3Utilities.get_s3_bucket_and_prefix_from_path(
                s3_target_file_full_path, s3_protocol_prefix)

            for obj in s3_resource.Bucket(s3_source_bucket).objects.filter(Prefix=s3_source_file_path_prefix):
                copy_source = {'Bucket': s3_source_bucket, 'Key': obj.key}
                new_obj_key = obj.key.replace(s3_source_file_path_prefix, s3_target_file_path_prefix)
                new_obj = s3_resource.Bucket(s3_target_bucket).Object(new_obj_key)
                new_obj.copy(copy_source)
                obj.delete()

        except Exception as e:
            raise Exception("Exception occurred while moving File, Error:%s, "
                            "s3_file_full_path:%s" % (str(e), s3_source_file_full_path))

    @staticmethod
    def s3_move_files_pending_to_processed(s3_file_full_path_list, s3_resource,  s3_protocol_prefix):

        for s3_file_full_path in s3_file_full_path_list:

            try:

                CommonS3Utilities.s3_move_file(s3_file_full_path,
                                               s3_file_full_path.replace('/pending/', '/processed/'),
                                               s3_resource, s3_protocol_prefix)
            except Exception as e:
                raise Exception("Exception occurred while moving File, Error:%s, "
                                "s3_file_full_path:%s" % (str(e), s3_file_full_path))

    @staticmethod
    def get_s3_bucket_and_prefix_from_path(s3_full_path, s3_protocol_prefix):

        s3_bucket, s3_path_prefix = s3_full_path.replace(s3_protocol_prefix, "").split("/", 1)
        return s3_bucket, s3_path_prefix




import argparse
import os
from .project_structure_generator import ProjectStructureGenerator, CommonLogging


def main():
    parser = argparse.ArgumentParser(
        description='Create a PySpark project structure template',
        usage='%(prog)s -name PROJECT_NAME'
    )
    parser.add_argument(
        '-name', '--name',
        required=True,
        help='Name of the project (must be at least three characters and a valid Python identifier)'
    )

    args = parser.parse_args()
    project_name = args.name
    
    # validate project name
    if not  (project_name  and len(project_name) >= 3 and project_name.replace('-', '_').isidentifier()):
        raise ValueError("Invalid Project Name: Project Name must be greater than or equal to three in length!")

    logging = CommonLogging.get_logger()
    try:
        logging.info(f"PySpark project structure generator '{project_name}'...")
        ProjectStructureGenerator.copy_and_rename_template(project_name)
        configs, full_config_file_path = ProjectStructureGenerator.read_config_file(project_name)
        configs = ProjectStructureGenerator.update_template_by_project_name(configs, full_config_file_path, project_name)
        ProjectStructureGenerator.write_to_config_file(configs, full_config_file_path)

        logging.info("Project structore creation complete.")
    except Exception as e:
        logging.error(f"Some thing went wrong while creating the project structure: {str(e)}")
        raise

if __name__ == "__main__":
    main()



import logging
from airflow.providers.cncf.kubernetes.operators.spark_kubernetes import SparkKubernetesOperator
from airflow.providers.cncf.kubernetes.sensors.spark_kubernetes import SparkKubernetesSensor
from airflow.utils.decorators import apply_defaults
from airflow.models import BaseOperator

# Creating custom Airflow spark Kubernetes operator to handle both submit and monitor
# ******************************************************************************************************


class CustomSparkKubernetesSubmitMonitor(BaseOperator):
    @apply_defaults
    def __init__(self, spark_kubernetes_operator_args, spark_kubernetes_sensor_args, global_vars, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.spark_kubernetes_operator_args = spark_kubernetes_operator_args
        self.spark_kubernetes_sensor_args = spark_kubernetes_sensor_args
        self.global_vars = global_vars

    def execute(self, context):
        spark_operator = SparkKubernetesOperator(task_id='spark-submit-task', **self.spark_kubernetes_operator_args)
        kubernetes_sensor = SparkKubernetesSensor(task_id='spark-monitor-task', **self.spark_kubernetes_sensor_args)
        try:
            spark_operator.execute(context)

        except Exception as e:
            if str(e).__contains__("AlreadyExists"):
                logging.warning("Spark Submit - Job already exists:%s" % str(e))
            else:
                logging.error("Spark Submit issue:%s" % str(e.with_traceback()))

                # Inserting Error info to Trino
                # self.insert_trino(context.get("task_instance").dag_id, context.get("task_instance").task_id, str(e))
                raise e

        try:
            kubernetes_sensor.execute(context)

        except Exception as e:
            logging.error("Spark Monitor issue:%s" % str(e))

            # *******
            # Inserting Error info to Trino
            # self.insert_trino(context.get("task_instance").dag_id, context.get("task_instance").task_id, str(e))
            raise e

    # ******************************************************************************************************
    # ******************************************************************************************************
    # Inserting error info to trino
    def insert_trino(self, dag_id, task_id, log_error_info):
        from datetime import datetime
        import re
        now = datetime.now()
        ingested_at = datetime.strptime(now.strftime("%Y-%m-%d %H:%M:%S"), "%Y-%m-%d %H:%M:%S")

        trino_statement = "INSERT INTO {} VALUES" \
                          " ('{}','{}', '{}', timestamp  '{}')".format(self.global_vars["airflow_error_details_table"],
                                                                       dag_id, task_id, log_error_info, ingested_at)

        try:
            from airflow.providers.trino.hooks.trino import TrinoHook
            trino_conn = TrinoHook(self.global_vars["trino_airflow_connections"]["datalake"])
            print(trino_conn)
            cursor = trino_conn.get_cursor()
            cursor.execute(trino_statement)
            response = cursor.fetchone()
            if response:
                print('> Error Details table insert has been completed successfully')
        except Exception as e:
            print('> Execution Failed')
            raise Exception(e, trino_statement)


# ******************************************************************************************************
# ******************************************************************************************************




import os
from airflow import DAG
# TODO: this is added to help airflow detect common_utils as module
import sys
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/../'))
print("path : {}".format(os.path.dirname(__file__) + '/../'))

from common_utils.flows import common_dag_module


# **************************  basic details  **************************************
local_workflows_config_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
                                           'config',
                                           os.path.basename(os.path.abspath(__file__)).replace(".py", ".json"))

pipeline_zone = ""
# **********************************************************************************


airflow_var_globals, airflow_var_pipelines, dag_default_args = \
    common_dag_module.DagCommons.read_airflow_variables(pipeline_zone)


configs, airflow_var_dag_details, dag_arguments = \
    common_dag_module.DagCommons.read_dag_details(local_workflows_config_path,
                                                  airflow_var_pipelines, dag_default_args, airflow_var_globals)

# DAG building task groups
with DAG(**dag_arguments) as dag:
    common_dag_module.DagCommons.build_task_group(dag, configs, airflow_var_globals, airflow_var_dag_details)


from pyspark.sql.functions import *
from bs4 import BeautifulSoup
from urllib.parse import urlencode, quote
import copy
from pyspark.sql.types import StructType
from pyspark.sql.functions import lit

# Custom Dependencies
# TODO: this is added to help airflow detect common_utils as module
import sys
import os
sys.path.append("/workflows/pyspark_template/")
# sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/../'))
# print("path : {}".format(os.path.dirname(__file__) + '/../'))

from common_utils.scripts.common_utilities import CommonLogging, CommonIcebergUtilities
from common_utils.scripts.helper.common_helper_utilities import CommonProcessHelperUtils, CommonAPIUtilities
from common_utils.scripts.data.common_data_utilities import CommonDataUtilities
from common_utils.scripts.s3.common_s3_utilities import CommonS3Utilities


logging = CommonLogging.get_logger()

# ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** *****

MOUNT_BUCKET = '/workflows'

class DefaultSparkCode:
    def get_dummy_data():
        data = [
            {
                "full_name":"John Smith",
                "dob":  "1980-02-09",
                "position": "Senior  Data Engineer",
                 "salary": 30000.00
            },
            {
                "full_name":"Hana Peter",
                "dob":  "1984-12-29",
                "position": "Software Engineer",
                 "salary": 25000.00
            },
            {
                "full_name":"Ali Abdela",
               "dob":  "1985-08-19",
                "position": "Data Engineer",
                 "salary": 28000.00
            }
        ]
        return data
    
   

    def ingest_data(s3_schema_file_path, schema_file_object_key,output_wh_table_load_strategy, 
                output_warehouse_fq_table, output_table_snapshots_delete_period):

        minio_s3_source = CommonS3Utilities.MinioS3Resource(minio_s3_credentials).minio_s3_resource

        config_table_schema = CommonProcessHelperUtils.get_schema_from_s3(s3_schema_file_path, schema_file_object_key, minio_s3_source)
        # config_table_schema =   CommonProcessHelperUtils.get_schema_from_local(s3_schema_file_path, schema_file_object_key)
        # get dummy data                                                             
        input_records_list =  DefaultSparkCode.get_dummy_data()     

        logging.info("config_spark_schema_json:" + str(config_table_schema))
        logging.info(f"Length of input_records_list:{str(len(input_records_list))}")

        data_rdd = sc.parallelize(input_records_list)

        format_data_rdd = data_rdd.map(lambda x: CommonDataUtilities.format_data(x, config_table_schema))

        spark_schema_json = CommonDataUtilities.create_spark_schema(config_table_schema)

        logging.info("spark_schema_json:" + str(spark_schema_json))
        data_df = spark.createDataFrame(format_data_rdd, schema=StructType.fromJson(spark_schema_json))

        # Adding this to remove duplicates, we may get duplicates
        data_df = data_df.drop_duplicates()

        data_df = data_df.withColumn("ingested_at", to_timestamp(lit(current_timestamp())))

        script_adding_columns = ["ingested_at"]
        logging.info(">> Dataframe schema: " + str(data_df.schema))
        logging.info(">> Dataframe spark_schema_json: " + str(spark_schema_json))
        logging.info(">> Display 10 rows: " + str(spark_schema_json))

        data_df.show(10, truncate=False)

        CommonIcebergUtilities.iceberg_load_operation(spark, output_wh_table_load_strategy, data_df,
                                                    output_warehouse_fq_table, script_adding_columns,
                                                    merge_data_join_fields=None)

        warehouse_catalog, warehouse_schema, warehouse_ref_table = output_warehouse_fq_table.split(".")

        # ********************************************************
        # Drop the data by ingested_at  -
        # Maintain only one snapshot in one week.
        # Maintain last 3 months data snapshots
        data_df.printSchema()
        # delete_sql_query = """
        #     DELETE FROM {} WHERE 
        #     (ingested_at < date_sub(CURRENT_TIMESTAMP(), {})) 
        #     OR (
        #             (ingested_at <> (SELECT max(ingested_at) FROM  {})) 
        #             AND 
        #             (ingested_at > date_sub(CURRENT_TIMESTAMP(), 6))
        #         )
        # """.format(output_warehouse_fq_table, output_table_snapshots_delete_period, output_warehouse_fq_table)

        # logging.info(f"Clear History snapshots, delete_sql_query:{delete_sql_query}")
        # spark.sql(delete_sql_query)
        # ********************************************************
        CommonIcebergUtilities.optimize_commands(spark, warehouse_catalog, warehouse_schema, warehouse_ref_table)
                                                             
    

def main():
    try:
        task_name = spark_job_args["task_name"]
        s3_schema_file_path = spark_job_args["s3_schema_file_path"]
        schema_file_object_key = spark_job_args["schema_file_object_key"]
        project_dir_name = spark_job_args["project_dir_name"]

        logging.info("""task_name: %s, s3_schema_file_path: %s, schema_file_object_key: %s
                """.format(task_name, s3_schema_file_path, schema_file_object_key))
        # Arguments - DW output table
        output_warehouse_fq_table = spark_job_args["output_warehouse_fq_table"]
        output_wh_table_load_strategy = spark_job_args["output_wh_table_load_strategy"]
        output_table_snapshots_delete_period = spark_job_args.get("output_table_snapshots_delete_period", 60)
        logging.info("""output_warehouse_fq_table: %s, output_wh_table_load_strategy: %s, output_table_snapshots_delete_period: %s
                """.format(output_warehouse_fq_table, output_wh_table_load_strategy, output_table_snapshots_delete_period))

        # Starter spark code  called
        DefaultSparkCode.ingest_data(s3_schema_file_path, schema_file_object_key,output_wh_table_load_strategy, 
                output_warehouse_fq_table, output_table_snapshots_delete_period)

      
    except Exception as ex:
        logging.error(ex)
        sc.stop()
        raise Exception('! Something went wrong with this job')

    finally:
        logging.info('Stopping...')
        sc.stop()


if __name__ == '__main__':
    # Provide Spark Session additional config options as dictionary
    spark_options = {
        "spark.sql.shuffle.partitions": "10",
        "spark.default.parallelism": "10"
    }
    spark, sc, spark_app_name, spark_job_args, cluster_details_spark_args, minio_s3_credentials,\
    credentials, *_ret_list = CommonProcessHelperUtils.initialize_read_arguments(spark_options=spark_options, mount_bucket_name=MOUNT_BUCKET)
    logging.info(f"spark_job_args: {spark_job_args}")
   

    main()
    exit()




from pyspark.sql.functions import col


class CommonLogging:
    @staticmethod
    def get_logger(log_level_str='INFO'):
        import logging
        try:
            level_num = logging.getLevelName(log_level_str.upper())
            if not isinstance(level_num, int):
                raise ValueError(f"Logging Level is not in predefined list: {str(level_num)}")

        except ValueError:
            level_num = logging.getLevelName('INFO')

        logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=level_num,
                            datefmt='%y/%m/%d %H:%M:%S')

        return logging.getLogger()


logging = CommonLogging.get_logger()


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


# will remove later to avoid breakage
class CommonAPIUtilities:
    pass


class CommonIcebergUtilities:

    @staticmethod
    def merge_upsert_process(spark, data_df, s3_warehouse_catalog, s3_warehouse_schema,
                             s3_warehouse_ref_table, join_fields):
        logging.info('> Merge data: Start')
        columns = data_df.columns
        new_data = data_df.select(columns)
        update_set_fields = list(set(columns) - set(join_fields))

        logging.info("update_set_fields:" + str(update_set_fields))

        # UPSERT
        new_data.createOrReplaceTempView("new_data")
        logging.info('> Columns: %s' % str(columns))

        # Having null validation , it may take time , so replace primary keys null values to empty spaces
        statement = \
            "MERGE INTO {}.{}.{} t USING new_data s ON {} \
                WHEN MATCHED THEN UPDATE SET {} \
                    WHEN NOT MATCHED THEN INSERT ({}) VALUES ({})".format(
                s3_warehouse_catalog, s3_warehouse_schema, s3_warehouse_ref_table,
                ' AND '.join(map(lambda jc: "((s.J_C=t.J_C) or (s.J_C IS NULL and t.J_C IS NULL))".replace("J_C", jc),
                                 join_fields)),
                ','.join(map(lambda f: 't.' + f + '=' + 's.' + f, update_set_fields)),
                ','.join(columns), ','.join(map(lambda c: 's.' + c, columns))
            )

        '''
        # Without null validation , Make sure to replace primary keys null values to empty spaces
        statement = \
            "MERGE INTO {}.{}.{} t USING new_data s ON {} \
                WHEN MATCHED THEN UPDATE SET {} \
                    WHEN NOT MATCHED THEN INSERT ({}) VALUES ({})".format(
                s3_warehouse_catalog, s3_warehouse_schema, s3_warehouse_ref_table,
                ' AND '.join(map(lambda jc: "(s.J_C=t.J_C)".replace("J_C", jc),
                                 join_fields)),
                ','.join(map(lambda f: 't.' + f + '=' + 's.' + f, update_set_fields)),
                ','.join(columns), ','.join(map(lambda c: 's.' + c, columns))
            )
        '''
        logging.info('Merge Statement: %s' % statement)
        spark.sql(statement).show(truncate=False)
        logging.info('> Merge data: Completed')

    @staticmethod
    def merge_insert_process(spark, data_df, s3_warehouse_catalog, s3_warehouse_schema,
                             s3_warehouse_ref_table, join_fields):
        logging.info('> Merge data: Start')
        columns = data_df.columns
        new_data = data_df.select(columns)

        # Merge insert
        new_data.createOrReplaceTempView("new_data")
        logging.info('> Columns: %s' % str(columns))

        statement = \
            "MERGE INTO {}.{}.{} t USING new_data s ON {} \
             WHEN NOT MATCHED THEN INSERT ({}) VALUES ({})".format(
                s3_warehouse_catalog, s3_warehouse_schema, s3_warehouse_ref_table,
                ' AND '.join(map(lambda jc: "((s.J_C IS NULL and t.J_C IS NULL) or (s.J_C=t.J_C))".replace("J_C", jc),
                                 join_fields)),
                ','.join(columns), ','.join(map(lambda c: 's.' + c, columns))
            )

        logging.info('Merge Statement: %s' % statement)
        spark.sql(statement).show(truncate=False)
        logging.info('> Merge insert data: Completed')

    @staticmethod
    def insert_overwrite_process(spark, data_df,
                                 s3_warehouse_catalog, s3_warehouse_schema, s3_warehouse_ref_table,
                                 warehouse_table_partition_details=None):
        logging.info('> Insert Overwrite data: Start')
        columns = data_df.columns
        new_data = data_df.select(columns)

        # Insert Overwrite
        new_data.createOrReplaceTempView("new_data")
        logging.info('> Columns: %s' % str(columns))

        if warehouse_table_partition_details is None:
            statement = \
                "INSERT OVERWRITE {}.{}.{} SELECT * from new_data".format(
                    s3_warehouse_catalog, s3_warehouse_schema, s3_warehouse_ref_table)

        else:
            statement = \
                "INSERT OVERWRITE {}.{}.{} PARTITION ({} = '{}') SELECT * from new_data".format(
                    s3_warehouse_catalog, s3_warehouse_schema, s3_warehouse_ref_table,
                    warehouse_table_partition_details["partition_column"],
                    warehouse_table_partition_details["partition_column_value"])

        logging.info('Insert Overwrite Statement: %s' % statement)
        spark.sql(statement).show(truncate=False)
        logging.info('> Insert Overwrite data: Completed')

    @staticmethod
    def delete_process(spark, data_df, s3_warehouse_catalog, s3_warehouse_schema,
                       s3_warehouse_ref_table, join_fields):
        logging.info('> Delete Record: Start')

        columns = data_df.columns
        new_data = data_df.select(columns)

        new_data.createOrReplaceTempView("new_data")
        logging.info('> Columns: %s' % str(columns))

        query_statement = \
            "MERGE INTO {}.{}.{} t USING new_data s ON {} \
                WHEN MATCHED THEN DELETE".format(
                s3_warehouse_catalog, s3_warehouse_schema, s3_warehouse_ref_table,
                ' AND '.join(map(lambda f: 's.' + f + '=' + 't.' + f, join_fields)))

        logging.info('Update Merge Statement: %s' % query_statement)
        spark.sql(query_statement).show(truncate=False)
        logging.info('> MERGE Delete Record: Completed')

    @staticmethod
    def optimize_commands(spark, catalog, schema, table):
        fq_table = "{}.{}.{}".format(catalog, schema, table)

        # ******************** ******************* ******************* ******************* *******************
        #             Debug                             #             Debug                               #

        logging.warning(" >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Starting- Before optimizations >>>>>>>>>>")
        logging.info(" >> >> Table History")
        ib_tbl_history = "SELECT * FROM {}.history".format(fq_table)
        spark.sql(ib_tbl_history).show(truncate=False)

        logging.info(" >> >> Data Files")
        ib_tbl_files = "SELECT * FROM {}.files".format(fq_table)
        spark.sql(ib_tbl_files).show(truncate=False)

        logging.info(" >> >> All Data Files")
        ib_tbl_all_data_files = "SELECT * FROM {}.all_data_files".format(fq_table)
        spark.sql(ib_tbl_all_data_files).show(truncate=False)

        logging.info(" >> >> Manifests Files")
        ib_tbl_manifest_files = "SELECT * FROM {}.manifests".format(fq_table)
        spark.sql(ib_tbl_manifest_files).show(truncate=False)

        logging.info(" >> >> All Manifests Files")
        ib_tbl_all_meta_files = "SELECT * FROM {}.all_manifests".format(fq_table)
        spark.sql(ib_tbl_all_meta_files).show(truncate=False)
        #             Debug                             #             Debug                               #
        # ******************** ******************* ******************* ******************* *******************

        logging.info(" >> >> Rewriting data files")
        ib_rw_data_files_query_stmt = "CALL {}.system.rewrite_data_files(table => '{}', " \
                                      "options => map('target-file-size-bytes','524288000'))".format(catalog, fq_table)
        spark.sql(ib_rw_data_files_query_stmt).show(truncate=False)

        logging.info(" >> >> Current snapshots before calling expire snapshots")
        spark.sql("SELECT committed_at, snapshot_id, operation FROM {}.snapshots".format(fq_table)).show(truncate=False)

        from datetime import datetime

        # ex_snp_tmp_str = datetime.today() - timedelta(days=1)
        ex_snp_tmp_str = datetime.today()
        ib_expire_snapshot_query_stmt = "CALL {}.system.expire_snapshots(table => '{}', older_than => TIMESTAMP '{}'," \
                                        " retain_last => 1, stream_results => true)".format(catalog,
                                                                                            fq_table, ex_snp_tmp_str)

        # ib_expire_snapshot_query_stmt = "CALL {}.system.expire_snapshots(table => '{}'," \
        #                                 " retain_last => 1, stream_results => true)".format(catalog, fq_table)

        try:

            spark.sql(ib_expire_snapshot_query_stmt).show(truncate=False)

        except Exception as e:
            logging.error(">> >>Expiry snapshot-exception in expiring snapshots: >> >> " + str(e))
            pass

        logging.info(" >> >> Current snapshots after calling expire snapshots")
        spark.sql("SELECT committed_at, snapshot_id, operation FROM {}.snapshots".format(fq_table)).show(truncate=False)

        logging.info(" >> >> Remove orphan files")
        ib_rm_orphan_files_query_stmt = "CALL {}.system.remove_orphan_files (table => '{}')".format(catalog, fq_table)
        spark.sql(ib_rm_orphan_files_query_stmt).show(truncate=False)

        # ******************** ******************* ******************* ******************* *******************
        #             Debug                             #             Debug                               #

        logging.warning(" >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> after remove orphan files >>>>>>>>>>>>>>>>>>>>>")
        logging.info(" >> >> Data Files")
        ib_tbl_files = "SELECT * FROM {}.files".format(fq_table)
        spark.sql(ib_tbl_files).show(truncate=False)

        logging.info(" >> >> All Data Files")
        ib_tbl_all_data_files = "SELECT * FROM {}.all_data_files".format(fq_table)
        spark.sql(ib_tbl_all_data_files).show(truncate=False)

        logging.info(" >> >> Manifests Files")
        ib_tbl_manifest_files = "SELECT * FROM {}.manifests".format(fq_table)
        spark.sql(ib_tbl_manifest_files).show(truncate=False)

        logging.info(" >> >> All Manifests Files")
        ib_tbl_all_meta_files = "SELECT * FROM {}.all_manifests".format(fq_table)
        spark.sql(ib_tbl_all_meta_files).show(truncate=False)

        #             Debug                             #             Debug                               #
        # ******************** ******************* ******************* ******************* *******************
        #

    @staticmethod
    def iceberg_load_operation(spark, output_wh_table_load_strategy, data_df, output_warehouse_fq_table,
                               script_adding_columns, merge_data_join_fields=None):

        table_catalog, table_schema, table = output_warehouse_fq_table.split(".")

        if output_wh_table_load_strategy.lower() == "insert_overwrite":
            logging.info(">> >> >> >> Loading table data - INSERT_OVERWRITE")

            # Re-ordering of columns as per table
            data_df = data_df.select(
                *list(CommonIcebergUtilities.get_tbl_column_info_select(spark, output_warehouse_fq_table).keys()))

            data_df.write.format("iceberg").mode("overwrite").save(output_warehouse_fq_table)

            # CommonIcebergUtilities.insert_overwrite_process(spark, data_df, table_catalog, table_schema, table, None)

        elif output_wh_table_load_strategy.lower() == "merge_insert":
            logging.info(">> >> >> >> Loading table data - MERGE_INSERT")
            if merge_data_join_fields is None:
                merge_data_join_fields = [x for x in data_df.columns if x not in script_adding_columns]

            CommonIcebergUtilities.merge_insert_process(spark, data_df, table_catalog, table_schema, table,
                                                        merge_data_join_fields)

        elif output_wh_table_load_strategy.lower() == "merge_upsert":
            logging.info(">> >> >> >> Loading table data - MERGE_UPSERT")
            if merge_data_join_fields is None:
                merge_data_join_fields = [x for x in data_df.columns if x not in script_adding_columns]

            data_df = data_df.sortWithinPartitions(*merge_data_join_fields)
            CommonIcebergUtilities.merge_upsert_process(spark, data_df, table_catalog, table_schema, table,
                                                        merge_data_join_fields)

        elif output_wh_table_load_strategy.lower() == "append":
            logging.info(">> >> >> >> Loading table data - APPEND")

            # Re-ordering of columns as per table
            data_df = data_df.select(
                *list(CommonIcebergUtilities.get_tbl_column_info_select(spark, output_warehouse_fq_table).keys()))

            data_df.writeTo(output_warehouse_fq_table).append()

        else:
            raise Exception(
                "Provide proper output_wh_table_load_strategy, provided as:%s" % output_wh_table_load_strategy)

    @staticmethod
    def get_tbl_column_info_describe(spark, output_warehouse_fq_table):
        # Getting columns and data type details of iceberg / warehouse table
        table_details_df = spark.sql("describe {}".format(output_warehouse_fq_table))
        table_details_df = table_details_df.filter((col("data_type").isNotNull()) & (col("data_type") != ""))
        table_details_df = table_details_df.select(*["col_name", "data_type"])
        table_details_dict_list = [row.asDict() for row in table_details_df.collect()]

        table_col_info = {}
        for table_details_dict in table_details_dict_list:
            table_col_info[table_details_dict["col_name"]] = table_details_dict["data_type"]

        return table_col_info

    @staticmethod
    def get_tbl_column_info_select(spark, output_warehouse_fq_table):

        table_cols_df = spark.sql("select * from {} limit 1".format(output_warehouse_fq_table))
        table_col_info = {}
        for col_info in table_cols_df.dtypes:
            table_col_info[col_info[0]] = col_info[1]

        return table_col_info



import argparse
import os
import shutil
import json

class Constants:
    RENAMING_ALLOWED_DIRECTORIES = ['flows', 'config', 'scripts']
    DEAULT_TEMPLATE_DIR_NAME = 'pyspark_template'
    DEAULT_TEMPLATE_DATA_SCHEMA_FILE_NAME = 'employee_data_schema'


class CommonLogging:
    @staticmethod
    def get_logger(log_level_str='INFO'):
        import logging
        try:
            level_num = logging.getLevelName(log_level_str.upper())
            if not isinstance(level_num, int):
                raise ValueError(f"Logging Level is not in predefined list: {str(level_num)}")

        except ValueError:
            level_num = logging.getLevelName('INFO')

        logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=level_num,
                            datefmt='%y/%m/%d %H:%M:%S')

        return logging.getLogger()


logging = CommonLogging.get_logger()

class ProjectStructureGenerator:
    @staticmethod
    def copy_and_rename_template(project_name):
        # Define paths
        template_dir = os.path.join(os.path.dirname(__file__), Constants.DEAULT_TEMPLATE_DIR_NAME)
        current_dir = os.getcwd()
        logging.info("current_dir: {}".format(current_dir))
        # Copy template directory to current directory
        try:
            shutil.copytree(template_dir, os.path.join(current_dir, project_name))
            logging.debug(f"Copying template directory from {template_dir} to {os.path.join(current_dir, project_name)}")
             # Log creation of first-level directories
            destination_dir = os.path.join(current_dir, project_name)
            for dirpath, _, _ in os.walk(destination_dir):
                if dirpath == destination_dir:
                    continue
                logging.info(f"Creating directory: {os.path.relpath(dirpath, destination_dir)}")

        except FileExistsError:
            logging.error(f"Error: Directory '{project_name}' already exists. Aborting.")
            return
        except Exception as e:
                raise Exception(f"Error occurred while copying template: {str(e)}")
        
        # List of directories where renaming is allowed
        for dir in Constants.RENAMING_ALLOWED_DIRECTORIES:
            dir_path = os.path.join(current_dir,project_name,  dir)
            if os.path.exists(dir_path):
                for filename in os.listdir(dir_path):
                    name, ext = os.path.splitext(filename)
                    if name in [Constants.DEAULT_TEMPLATE_DIR_NAME, Constants.DEAULT_TEMPLATE_DATA_SCHEMA_FILE_NAME] and ext != '':
                        # Determine new filename with the project_name preserved extension
                        new_filename = project_name + ext
                        if name.endswith("_schema") and ext == ".json":
                            new_filename = project_name.replace('-', '_') + "_schema.json"

                        original_file_path = os.path.join(dir_path, filename)
                        new_file_path = os.path.join(dir_path, new_filename)
                        os.rename(original_file_path, new_file_path)
                        logging.debug(f"Renamed {filename} to {new_filename}")


    @staticmethod
    def read_config_file(project_name):
        current_dir = os.getcwd()
        file_name = project_name + '.json'
        full_config_file_path = os.path.join(current_dir, project_name, "config", file_name)
        try:
                with open(full_config_file_path, "r") as f:
                    configurations = json.loads(f.read())

                return configurations, full_config_file_path
        except Exception as e:
            raise Exception( "Exception occurred while reading Config File: %s, Error:%s " % (full_config_file_path, str(e)))
    @staticmethod
    def write_to_config_file(configs, full_config_file_path):
        try:
            with open(full_config_file_path, "w") as f:
                configs_json = json.dumps(configs, indent=2)
                f.write(configs_json)
                logging.debug(f"Writing updates to config file completed.")
        except Exception as e:
                raise Exception( "Exception occurred while updating Config File: %s, Error:%s " % (full_config_file_path, str(e)))

    @staticmethod
    def update_template_by_project_name(configs, full_config_file_path, project_name):
        aiflow_dag_details = configs["aiflow-dag-details"]
        aiflow_dag_details["dag_id"] = project_name.replace('-', '_')

        airflow_task_details = configs.get("airflow-task-details", {})
        if airflow_task_details is not None:
            project_name_job = project_name.replace('_', '-')
            airflow_task_details[project_name_job] = airflow_task_details["pyspark-template"]
            
            pyspark_job_main_file_s3_path = airflow_task_details[project_name_job]["spark_details"]["spark_kubernetes_args"]["pyspark_job_main_file_s3_path"]
            update_pyspark_job_main_file_s3_path = pyspark_job_main_file_s3_path.replace(Constants.DEAULT_TEMPLATE_DIR_NAME, project_name)
            airflow_task_details[project_name_job]["spark_details"]["spark_kubernetes_args"]["pyspark_job_main_file_s3_path"] = update_pyspark_job_main_file_s3_path
            #Delete keys except the project_name_job
            keys_to_remove = [key for key in airflow_task_details.keys() if key != project_name_job]
            for key in keys_to_remove:
                del airflow_task_details[key]
            # update schema dir path
            old_s3_schema_file_path =  airflow_task_details[project_name_job]["spark_details"]["spark_job_args"]["s3_schema_file_path"]
            new_s3_schema_file_path = old_s3_schema_file_path.replace(Constants.DEAULT_TEMPLATE_DIR_NAME, project_name)
            # update file name
            new_s3_schema_file_path = new_s3_schema_file_path.replace((Constants.DEAULT_TEMPLATE_DATA_SCHEMA_FILE_NAME + ".json"), (project_name.replace('-', '_') + "_schema.json"))
            airflow_task_details[project_name_job]["spark_details"]["spark_job_args"]["s3_schema_file_path"] = new_s3_schema_file_path
            # add project_dir_name, taks_name, schema_file_object_key
            airflow_task_details[project_name_job]["spark_details"]["spark_job_args"]["project_dir_name"] = project_name
            airflow_task_details[project_name_job]["spark_details"]["spark_job_args"]["task_name"] = project_name
            # airflow_task_details[project_name_job]["spark_details"]["spark_job_args"]["schema_file_object_key"] = project_name

             
            configs["airflow-task-details"] = airflow_task_details
            configs["aiflow-dag-details"] = aiflow_dag_details
            
        return configs
     



from datetime import datetime, timedelta
from airflow.models import Variable
from airflow.operators.dummy import DummyOperator as DummyOperator
from airflow.operators.python import PythonOperator
from airflow.utils.task_group import TaskGroup
from airflow.datasets import Dataset
import yaml
# from airflow.providers.cncf.kubernetes.operators.spark_kubernetes import SparkKubernetesOperator
# from airflow.providers.cncf.kubernetes.sensors.spark_kubernetes import SparkKubernetesSensor
# from airflow.utils.decorators import apply_defaults
# from airflow.models import BaseOperator


class Constants:
    DIR_AIRFLOW_DAG_ENV = "/opt/airflow/dags"
    DIR_AIRFLOW_LOG_ENV = "/opt/airflow/logs"
    S3_PROTOCOL_PREFIX = "s3a://"
    AIRFLOW_TRIGGER_ALL_SUCCESS = "all_success"
    # TODO: DEFAULT_AIRFLOW_HTTPS_PROXY is added for testing
    DEFAULT_AIRFLOW_HTTPS_PROXY = ""
    # provide default list of dependencies list
    # TODO: DEFAULT_IMPORT_STATEMENTS_LIST is empty
    DEFAULT_IMPORT_STATEMENTS_LIST = ""
    # DEFAULT_IMPORT_STATEMENTS_LIST = [f"from transformed.common_utils.dependencies.scripts import common_utilities"]
    # Provide default Python operator callable
      # TODO: DEFAULT_PYTHON_CALLABLE is empty
    DEFAULT_PYTHON_CALLABLE = ""
    # DEFAULT_PYTHON_CALLABLE = "common_utilities.ProcessorUtils.main_trino_process"

    # provide default Spark main script S3 path
    # TODO: DEFAULT_PYSPARK_JOB_MAIN_FILE_S3_PATH is empty since starter is provided under root_project/scripts/
    DEFAULT_PYSPARK_JOB_MAIN_FILE_S3_PATH = ""
   
    # Provide comma separated local paths of Spark (Mounted works volume)
    # TODO: DEFAULT_PYSPARK_DEPENDENCIES_LOCAL_PATH is empty 
    DEFAULT_PYSPARK_DEPENDENCIES_LOCAL_PATH = ""

class KubernetesSparkSetup:

    # ***** ***** Pyspark Kubernetes Initialization ***** ***** ***** *****

    @staticmethod
    def handle_min_resources(spark_kubernetes_job_args):
        driver_cores = spark_kubernetes_job_args["driver_cores"]
        executor_cores = spark_kubernetes_job_args["executor_cores"]
        driver_memory = str(spark_kubernetes_job_args["driver_memory"])
        executor_memory = str(spark_kubernetes_job_args["executor_memory"])
        executor_instances = spark_kubernetes_job_args["executor_instances"]

        if str(driver_memory).__contains__("g"):
            d_memory = int(driver_memory.split("g")[0]) * 1024
        elif str(driver_memory).__contains__("m"):
            d_memory = int(driver_memory.split("m")[0])
        else:
            raise Exception("Provide driver_memory in as suffix as 'g' or 'm'")

        if str(executor_memory).__contains__("g"):
            e_memory = int(executor_memory.split("g")[0]) * 1024
        elif str(executor_memory).__contains__("m"):
            e_memory = int(executor_memory.split("m")[0])
        else:
            raise Exception("Provide executor_memory in as suffix as 'g' or 'm'")

        min_cores = str(driver_cores + (executor_instances * executor_cores))
        min_memory = str(d_memory + (executor_instances * e_memory)) + "m"

        return min_cores, min_memory

    @staticmethod
    def create_cluster_details_spark_args(airflow_var_globals, airflow_var_dag_details):
        cluster_details_spark_args = {
            "s3_credentials": {
                "s3_host": airflow_var_globals["s3_host"],
                 "s3_access_key": airflow_var_globals["s3_access_key"],
                "s3_secret_access_key": airflow_var_globals["s3_secret_access_key"],
                "s3_region": airflow_var_globals["s3_region"],
                "s3_signature_version": airflow_var_globals["s3_signature_version"]
            },
             "https_proxy": airflow_var_globals.get("https_proxy", Constants.DEFAULT_AIRFLOW_HTTPS_PROXY),
            "catalog": airflow_var_dag_details.get("catalog", None),
            "catalog_minio_bucket": airflow_var_globals["buckets"][
                airflow_var_dag_details["catalog"]] if airflow_var_dag_details.get("catalog",
                                                                                   None) is not None else None,
            "global_vars": airflow_var_globals,
             "airflow_var_dag_details": airflow_var_dag_details
        }

        return cluster_details_spark_args

    @staticmethod
    def update_default_values_spark_kubernetes_job_args(spark_kubernetes_job_args):
        # Spark code language
        if 'spark_language' not in spark_kubernetes_job_args:
            spark_kubernetes_job_args['spark_language'] = 'python'

        # Spark volcano scheduler default property
        if 'volcano_job_priority' not in spark_kubernetes_job_args:
            spark_kubernetes_job_args['volcano_job_priority'] = 'routine'

        # Spark driver resources
        if 'driver_cores' not in spark_kubernetes_job_args:
            spark_kubernetes_job_args['driver_cores'] = 1
        if 'driver_memory' not in spark_kubernetes_job_args:
            spark_kubernetes_job_args['driver_memory'] = '4g'

        # Spark Executor resources
        if 'executor_instances' not in spark_kubernetes_job_args:
            spark_kubernetes_job_args['executor_instances'] = 1
        if 'executor_cores' not in spark_kubernetes_job_args:
            spark_kubernetes_job_args['executor_cores'] = 1
        if 'executor_memory' not in spark_kubernetes_job_args:
            spark_kubernetes_job_args['executor_memory'] = '4g'

        if 'spark_dependencies_libraries' not in spark_kubernetes_job_args:
            spark_kubernetes_job_args['spark_dependencies_libraries'] = ""

        if "pyspark_job_main_file_s3_path" not in spark_kubernetes_job_args:
            spark_kubernetes_job_args["pyspark_job_main_file_s3_path"] = Constants.DEFAULT_PYSPARK_JOB_MAIN_FILE_S3_PATH

        return spark_kubernetes_job_args

    @staticmethod
    def update_default_values_spark_job_args(spark_job_args):
        if "python_dependencies_base_paths" not in spark_job_args:
            spark_job_args["python_dependencies_base_paths"] = Constants.DEFAULT_PYSPARK_DEPENDENCIES_LOCAL_PATH

        return spark_job_args
     
    @staticmethod
    def build_k8s_pyspark_configuration(spark_app_name, spark_kubernetes_job_args, spark_job_args, airflow_var_globals,
                                        airflow_var_dag_details, job_arguments=[]):
        import json

        # Update default values for Spark job submission on Kubernetes
        spark_kubernetes_job_args = KubernetesSparkSetup.\
            update_default_values_spark_kubernetes_job_args(spark_kubernetes_job_args)

        # Update default values for Spark job Arguments
        spark_job_args = KubernetesSparkSetup.\
            update_default_values_spark_job_args(spark_job_args)

        spark_language = spark_kubernetes_job_args['spark_language'].lower()

        cluster_details_spark_args = KubernetesSparkSetup.create_cluster_details_spark_args(airflow_var_globals,
                                                                                            airflow_var_dag_details)
        hadoop_custom_arguments = {
            "spark_job_args": spark_job_args, "cluster_details_spark_args": cluster_details_spark_args
        }
        # TODO: below is added to make the catalog and it's bucket dynamic
        catalog = cluster_details_spark_args["catalog"] if cluster_details_spark_args["catalog"] else None
        k8s_configuration = {
            "apiVersion": "sparkoperator.k8s.io/v1beta2", "kind": "SparkApplication", "metadata": {
                "name": spark_app_name, "namespace": airflow_var_globals["k8s_namespace"],
                "labels": {"app": "spark-jobs"}
            }, "spec": {

                # "deps": {

                # "pyFiles": [
                # "https: "
                # ]
                # },

                "type": "Scala" if spark_language == 'scala' else 'Python',
                 "mode": "cluster",
                "image": airflow_var_globals["spark_image"] if spark_language == 'scala' else airflow_var_globals["pyspark_image"],
                "imagePullPolicy": "Always",
                "mainApplicationFile": spark_kubernetes_job_args["pyspark_job_main_file_s3_path"],
                "arguments": job_arguments, "sparkVersion": airflow_var_globals["spark_version"],
                "batchScheduler": "volcano",
                "batchSchedulerOptions": {
                    "priorityClassName": spark_kubernetes_job_args['volcano_job_priority'],
                    # routine for normal , rush for priority jobs
                    # "minMember": 55,
                    "resources": {
                        "cpu": KubernetesSparkSetup.handle_min_resources(spark_kubernetes_job_args)[0],
                        "memory": KubernetesSparkSetup.handle_min_resources(spark_kubernetes_job_args)[1]
                    }
                }, "restartPolicy": {
                    "type": "Never"
                }, "timeToLiveSeconds": airflow_var_globals["k8s_pod_ttl"],

                "volumes": [
                        {
                            "name": airflow_var_globals["volume_details"]["volumes_name"],
                            "persistentVolumeClaim": {
                                "claimName": airflow_var_globals["volume_details"]["volumes_claim_name"]
                                }
                        }
                        ,
                        {
                            "name": airflow_var_globals["volume_details"]["ca_bundle_volume_name"],
                            "configMap": {
                                "name": airflow_var_globals["volume_details"]["ca_bundle_volume_name"],
                                "items": [
                                    {
                                        "key": f"{airflow_var_globals['volume_details']['ca_bundle_volume_name']}.crt",
                                        "path": airflow_var_globals["volume_details"]["ca_bundle_volume_mount_sub_path"]
                                    }
                                ]
                            }
                        },
                        {
                            "name": airflow_var_globals["volume_details"]["key_store_volume_name"],
                            "configMap": {
                                "name": airflow_var_globals["volume_details"]["ca_bundle_volume_name"],
                                "items": [
                                    {
                                        "key":  airflow_var_globals["volume_details"]["key_store_volume_mount_sub_path"],
                                        "path":  airflow_var_globals["volume_details"]["key_store_volume_mount_sub_path"]
                                    }
                                ]
                            }
                        }


                ],
                "driver": {
                    "cores": spark_kubernetes_job_args["driver_cores"],  # "coreLimit": "1200m",
                    "coreLimit": str(spark_kubernetes_job_args["driver_cores"]),
                    "memory": spark_kubernetes_job_args["driver_memory"], "labels": {"version": "3.1.1"},
                    "serviceAccount": airflow_var_globals["k8s_service_account"],

                    "volumeMounts": [{
                        "mountPath": airflow_var_globals["volume_details"]["volume_mounts_mount_path"],
                        "name": airflow_var_globals["volume_details"]["volume_mounts_name"]
                    }
                      ,  

                    {
                        "mountPath": airflow_var_globals["volume_details"]["ca_bundle_volume_mount_path"],
                        "subPath": airflow_var_globals["volume_details"]["ca_bundle_volume_mount_sub_path"],
                        "name": airflow_var_globals["volume_details"]["ca_bundle_volume_name"]
                    }, {
                            "mountPath": airflow_var_globals["volume_details"]["key_store_volume_mount_path"],
                            "subPath": airflow_var_globals["volume_details"]["key_store_volume_mount_sub_path"],
                            "name": airflow_var_globals["volume_details"]["key_store_volume_name"]
                        }
                    ],
                     "envVars": {
                       "JAVA_OPTS": f"$JAVA_OPTS -Djavax.net.ssl.trustStore= {airflow_var_globals['volume_details']['key_store_volume_mount_path']}",
                        "REQUESTS_CA_BUNDLE": airflow_var_globals["volume_details"]["ca_bundle_volume_mount_path"]

                    }
                    

                },

                "executor": {
                    "instances": spark_kubernetes_job_args["executor_instances"],
                    "cores": spark_kubernetes_job_args["executor_cores"],
                    "coreLimit": str(spark_kubernetes_job_args["executor_cores"]),
                    "memory": spark_kubernetes_job_args["executor_memory"], "labels": {"version": "3.1.1"},
                    "volumeMounts": [
                        {
                        "mountPath": airflow_var_globals["volume_details"]["volume_mounts_mount_path"],
                        "name": airflow_var_globals["volume_details"]["volume_mounts_name"]
                    }
                    ,  

                    {
                        "mountPath": airflow_var_globals["volume_details"]["ca_bundle_volume_mount_path"],
                        "subPath": airflow_var_globals["volume_details"]["ca_bundle_volume_mount_sub_path"],
                        "name": airflow_var_globals["volume_details"]["ca_bundle_volume_name"]
                    }, {
                            "mountPath": airflow_var_globals["volume_details"]["key_store_volume_mount_path"],
                            "subPath": airflow_var_globals["volume_details"]["key_store_volume_mount_sub_path"],
                            "name": airflow_var_globals["volume_details"]["key_store_volume_name"]
                        }
                    ],
                     "envVars": {
                       "JAVA_OPTS": f"$JAVA_OPTS -Djavax.net.ssl.trustStore= {airflow_var_globals['volume_details']['key_store_volume_mount_path']}",
                        "REQUESTS_CA_BUNDLE": airflow_var_globals["volume_details"]["ca_bundle_volume_mount_path"]

                    }

                }, "sparkConf": {
                    "spark.hadoop.custom.arguments": json.dumps(hadoop_custom_arguments, default=str),
                    "spark.sql.extensions": "org.apache.iceberg.spark.extensions.IcebergSparkSessionExtensions",
                    "spark.sql.catalog.spark_catalog": "org.apache.iceberg.spark.SparkSessionCatalog",
                    "spark.sql.catalog.spark_catalog.type": "hive",
                    "spark.sql.catalog.datawarehouse": "org.apache.iceberg.spark.SparkCatalog",
                    "spark.sql.catalog.datawarehouse.uri": airflow_var_globals["metastores"]["datawarehouse"],
                    "spark.sql.iceberg.handle-timestamp-without-timezone": "true", "spark.eventLog.enabled": "true",
                    "spark.history.fs.logDirectory": airflow_var_globals["s3_spark_logging_path"],
                    "spark.eventLog.dir": airflow_var_globals["s3_spark_logging_path"],
                    "spark.hadoop.fs.s3a.endpoint": airflow_var_globals["s3_host"],
                    "spark.hadoop.fs.s3a.access.key": airflow_var_globals["s3_access_key"],
                    "spark.hadoop.fs.s3a.secret.key": airflow_var_globals["s3_secret_access_key"],
                    "spark.hadoop.fs.s3a.connection.ssl.enabled": "true",
                    "spark.hadoop.fs.s3a.path.style.access": "true",
                    "spark.executor.extraClassPath": spark_kubernetes_job_args["spark_dependencies_libraries"],
                    "spark.driver.extraClassPath": spark_kubernetes_job_args["spark_dependencies_libraries"],
                    "spark.sql.parquet.fs.optimized.committer.optimization-enabled": "true",
                    "spark.hadoop.fs.s3a.connection.maximum": "100",
                    "spark.ui.proxyRedirectUri": airflow_var_globals["spark_ui_proxy_redirect_uri"]
                }
            }
        }
        #TODO: Check if the user used different catalog for the dag id outside "datawarehouse"
        if catalog is not None and catalog not in ["datawarehouse"]:
            # add the catalog to the sparkConf
            k8s_configuration["spec"]["sparkConf"][f"spark.sql.catalo.{catalog}"] = "org.apache.iceberg.spark.SparkCatalog"
            k8s_configuration["spec"]["sparkConf"][f"spark.sql.catalo.{catalog}.url"]= airflow_var_globals["metastores"][catalog]

        return k8s_configuration


class BuildDagTaskGroup:

    @staticmethod
    def build_spark_details(dag, task_group_key, task_details_key, airflow_task_details, airflow_var_globals,
                            airflow_var_dag_details, common_arguments):
        # Custom spark operator import
        from common_utils.flows.custom_operator import CustomSparkKubernetesSubmitMonitor
        airflow_tasks_list = []
        spark_kubernetes_args = airflow_task_details[task_group_key][task_details_key].get("spark_kubernetes_args", {})
        spark_job_args = airflow_task_details[task_group_key][task_details_key].get("spark_job_args", {})

        # Adding Common Arguments to Spark Arguments
        for common_args_key in common_arguments.keys():
            if common_args_key not in spark_job_args.keys():
                spark_job_args[common_args_key] = common_arguments[common_args_key]

        # ********* getting meta store details *********************************************

        spark_submit_application_config = KubernetesSparkSetup.build_k8s_pyspark_configuration(
            "{}".format(task_group_key), spark_kubernetes_args, spark_job_args, airflow_var_globals,
            airflow_var_dag_details,
            job_arguments=["{}".format(task_group_key)])
        app_file  = yaml.dump(spark_submit_application_config)
        spark_kubernetes_operator_args = {
            "namespace": airflow_var_globals["k8s_namespace"],
            "kubernetes_conn_id": airflow_var_globals["k8s_connection_id"],
            "application_file": app_file
        }
        spark_kubernetes_sensor_args = {
            "namespace": airflow_var_globals["k8s_namespace"],
            "kubernetes_conn_id": airflow_var_globals["k8s_connection_id"],
            "application_name": task_group_key, "attach_log": True
        }
        spark_job = CustomSparkKubernetesSubmitMonitor(task_id="{}".format(task_group_key),
                                                       spark_kubernetes_operator_args=spark_kubernetes_operator_args,
                                                       spark_kubernetes_sensor_args=spark_kubernetes_sensor_args,
                                                       global_vars=airflow_var_globals,
                                                       dag=dag)
        airflow_tasks_list.append(spark_job)

        return airflow_tasks_list

    @staticmethod
    def build_python_details(dag, task_group_key, task_details_key, airflow_task_details, global_vars,
                             airflow_var_dag_details):

        python_callable_args = airflow_task_details[task_group_key][task_details_key]
        import_statements_list = python_callable_args.get("import_statements_list",
                                                          Constants.DEFAULT_IMPORT_STATEMENTS_LIST)
        python_callable = python_callable_args.get("python_callable", Constants.DEFAULT_PYTHON_CALLABLE)

        common_arguments = {"DIR_AIRFLOW_DAG_ENV": Constants.DIR_AIRFLOW_DAG_ENV}

        import importlib

        for import_statement in import_statements_list:
            try:
                exec(import_statement)
            except Exception as e:
                print(f"Error Executing Import import_statement - {import_statement}: {e}")

        airflow_tasks_list = []

        python_task = PythonOperator(task_id=task_group_key,
                                     python_callable=eval(python_callable),
                                     op_args=[python_callable_args, global_vars, airflow_var_dag_details,
                                              common_arguments],
                                     dag=dag)

        airflow_tasks_list.append(python_task)
        return airflow_tasks_list


class DagCommons:

    @staticmethod
    def format_boolean_value(boolean_value, default_value):
        if boolean_value is not None and isinstance(boolean_value, bool) and boolean_value is False:
            task_status_mail_flag = False
        elif boolean_value is not None and isinstance(boolean_value, bool) and boolean_value is True:
            task_status_mail_flag = True
        else:
            task_status_mail_flag = default_value
        return task_status_mail_flag

    @staticmethod
    def read_airflow_variables(pipeline_zone):
        airflow_var_globals = Variable.get("GLOBALS", deserialize_json=True)

        if pipeline_zone is None or pipeline_zone == "":
            airflow_variable_suffix = ""
        else:
            airflow_variable_suffix = f"_{pipeline_zone.upper()}"

        airflow_var_pipelines = Variable.get(f"AIRFLOW_PIPELINES{airflow_variable_suffix}",
                                             deserialize_json=True)

        dag_default_args = Variable.get(f"AIRFLOW_DAG_DEFAULT_ARGS{airflow_variable_suffix}",
                                        deserialize_json=True)

        return airflow_var_globals, airflow_var_pipelines, dag_default_args

    @staticmethod
    def read_config_from_local(local_workflows_config_path):
        import json
        try:
            with open(local_workflows_config_path, "r") as f:
                configurations = json.loads(f.read())

            return configurations
        except Exception as e:
            raise Exception(
                "Exception occurred while reading Config File: %s, Error:%s " % (local_workflows_config_path, str(e)))

    @staticmethod
    def update_task_retries_and_retry_delay(airflow_var_dag_details, dag_default_args):

        retries_dict = {}
        if "retries" in airflow_var_dag_details and airflow_var_dag_details["retries"] != "" and \
                airflow_var_dag_details["retries"] is not None:

            dag_default_args["retries"] = airflow_var_dag_details["retries"]

        elif "retries" not in dag_default_args:
            dag_default_args["retries"] = 1

        if "retry_delay" in airflow_var_dag_details and airflow_var_dag_details["retry_delay"] != "" and \
                airflow_var_dag_details["retry_delay"] is not None:

            retries_dict["retry_delay"] = timedelta(minutes=airflow_var_dag_details["retry_delay"])

        # retry_delay default value is 5 minutes and if we pass it as variable, then it will be in seconds = 900
        return dag_default_args

    # @staticmethod
    # def update_task_mail_alerts(airflow_var_dag_details, dag_default_args, airflow_var_globals):

    #     task_status_mail_flag = DagCommons.format_boolean_value(airflow_var_dag_details.get("task_status_mail_flag"),
    #                                                             True)

    #     if task_status_mail_flag:

    #         from common_utils.flows.email_alert_notification import EmailAlert
    #         on_failure_callback = EmailAlert.on_failure_callback(dag_default_args.get('email_to'),
    #                                                              airflow_var_globals.get('airflow_url'))
    #         on_retry_callback = EmailAlert.on_retry_callback(dag_default_args.get('email_to'),
    #                                                          airflow_var_globals.get('airflow_url'))
    #         on_success_callback = EmailAlert.on_success_callback(dag_default_args.get('email_to'),
    #                                                              airflow_var_globals.get('airflow_url'))

    #         dag_default_args.update({'on_failure_callback': on_failure_callback})
    #         dag_default_args.update({'on_retry_callback': on_retry_callback})
    #         dag_default_args.update({'on_success_callback': on_success_callback})

    #     return dag_default_args

    # Handle Dag Arguments
    @staticmethod
    def generate_dag_arguments(airflow_dag_details_from_config, airflow_var_dag_details, dag_default_args):
        dag_arguments = {
            "dag_id": airflow_dag_details_from_config["dag_id"],  # "default_args": {'max_active_runs': 1},
            "default_args": dag_default_args,
            "description": airflow_dag_details_from_config.get("dag_description", ""),
            "schedule_interval": airflow_var_dag_details.get("schedule_interval", None),
            "start_date": airflow_var_dag_details["start_date"],
            "catchup": False,
            "tags": airflow_var_dag_details.get("tags", None)
        }
        return dag_arguments

    @staticmethod
    def update_airflow_dag_schedule_details(airflow_var_dag_details):
        schedule_interval = airflow_var_dag_details.get("schedule_interval", None)
        if schedule_interval == "":
            airflow_var_dag_details["schedule_interval"] = None
        else:
            airflow_var_dag_details["schedule_interval"] = schedule_interval

        start_date = airflow_var_dag_details.get("start_date", "")

        if start_date == "" or start_date is None:
            now = datetime.now() - timedelta(days=30)
            airflow_var_dag_details["start_date"] = datetime.strptime(now.strftime("%Y-%m-%d %H:%M:%S"),
                                                                      "%Y-%m-%d %H:%M:%S")
        else:
            airflow_var_dag_details["start_date"] = datetime.strptime(str(start_date), "%Y-%m-%d %H:%M:%S")

        return airflow_var_dag_details

    #  Not using currently , will use after refactoring the code
    @staticmethod
    def read_dag_details(workflows_config_path, airflow_var_pipelines, dag_default_args, airflow_var_globals):
        # Calling common dag module and building the task group

        configs = DagCommons.read_config_from_local(workflows_config_path)

        airflow_dag_details_from_config = dict(configs["aiflow-dag-details"])

        airflow_var_dag_details = airflow_var_pipelines.get(airflow_dag_details_from_config["dag_id"], {})

        airflow_var_dag_details = DagCommons.update_airflow_dag_schedule_details(airflow_var_dag_details)

        # Get airflow DAG Retries
        dag_default_args = DagCommons.update_task_retries_and_retry_delay(airflow_var_dag_details, dag_default_args)
        # TODO: update_task_mail_alerts connected out temporarly 
        # dag_default_args = DagCommons.update_task_mail_alerts(airflow_var_dag_details, dag_default_args,
                                                            #   airflow_var_globals)

        dag_arguments = DagCommons.generate_dag_arguments(airflow_dag_details_from_config, airflow_var_dag_details,
                                                          dag_default_args)

        return configs, airflow_var_dag_details, dag_arguments

    @staticmethod
    def build_task_group(dag, configs, airflow_var_globals, airflow_var_dag_details):
        # ***** *****  Reading Job configuration parameters ***** ***** ***** *****

        airflow_task_details = configs["airflow-task-details"]
        aiflow_task_dependencies_details = configs.get("aiflow-task-dependencies-details", {})
        common_arguments = configs.get("common-arguments", {})

        airflow_tasks_groups_list = []
        for task_group_key in airflow_task_details.keys():

            tasks_list = []

            for task_details_key in airflow_task_details[task_group_key].keys():
                # ########################### Spark tasks ##########################
                if task_details_key == "spark_details":
                    tasks_list = BuildDagTaskGroup.build_spark_details(dag, task_group_key, task_details_key,
                                                                       airflow_task_details, airflow_var_globals,
                                                                       airflow_var_dag_details, common_arguments)

                elif task_details_key == "python_details":
                    tasks_list = BuildDagTaskGroup.build_python_details(dag, task_group_key, task_details_key,
                                                                        airflow_task_details, airflow_var_globals,
                                                                        airflow_var_dag_details)

            if len(tasks_list) > 1:
                with TaskGroup(group_id=task_group_key) as tg:
                    # Building task dependencies inside task group (sequential)
                    for j in range(0, len(tasks_list)):
                        if j != 0:
                            tasks_list[j].set_upstream(tasks_list[j - 1])

                    airflow_tasks_groups_list.append((tg, task_group_key))
            else:
                for task in tasks_list:
                    airflow_tasks_groups_list.append((task, task_group_key))

        DagCommons.build_task_dependencies(dag, airflow_tasks_groups_list, aiflow_task_dependencies_details,
                                           airflow_var_dag_details)

    @staticmethod
    def build_task_dependencies(dag, airflow_tasks_groups_list, aiflow_task_dependencies_details,
                                airflow_var_dag_details):
        all_nodes_list = []
        dependency_tuples_list = []
        added_nodes_dict = {}

        all_nodes_list.extend(airflow_tasks_groups_list)

        dataset_outlet_flag = DagCommons.format_boolean_value(airflow_var_dag_details.get("dataset_outlet_flag"), False)
        end_outlets_kwargs = {"outlets": [Dataset(str(dag.dag_id))] if dataset_outlet_flag else {}}

        if "start" not in dag.tasks:
            task = DummyOperator(task_id="start", dag=dag)
            all_nodes_list.insert(0, (task, "start"))
            added_nodes_dict["start"] = task

        if "end" not in dag.tasks:
            # Data set triggers at end task
            task = DummyOperator(task_id="end", dag=dag, **end_outlets_kwargs)
            all_nodes_list.append((task, "end"))
            added_nodes_dict["end"] = task

        functional_nodes_count = len(airflow_tasks_groups_list)
        functional_nodes_list = airflow_tasks_groups_list

        def build_sequential_dependencies():

            for i in range(0, len(all_nodes_list) - 1):
                current_node = all_nodes_list[i][0]
                next_node = all_nodes_list[i + 1][0]
                dependency_tuples_list.append((current_node, next_node))

        def build_parallel_dependencies(task_auto_dependencies={}):

            max_tasks_per_group = task_auto_dependencies.get("max_tasks_per_group",
                                                             functional_nodes_count)
            trigger_rule = task_auto_dependencies.get("trigger_rule", "all_success")

            # Generating dummy nodes
            group_nodes_dict = {}

            group_no_list = []
            for i in range(0, functional_nodes_count):

                group_no = int(i / max_tasks_per_group)
                if group_no not in group_no_list and group_no < int((functional_nodes_count - 1) / max_tasks_per_group):
                    group_nodes_dict[f"I{group_no + 1}"] = DummyOperator(task_id=f"I{group_no + 1}", dag=dag,
                                                                         trigger_rule=trigger_rule)
                    group_no_list.append(group_no)

            # Building relationship
            for i in range(0, len(functional_nodes_list)):
                current_node = functional_nodes_list[i][0]

                group_no = int(i / max_tasks_per_group)

                if group_no == 0:
                    predecessor_node = added_nodes_dict["start"]
                else:
                    predecessor_node = group_nodes_dict[f"I{group_no}"]

                if group_no < int((functional_nodes_count - 1) / max_tasks_per_group):
                    successor_node = group_nodes_dict[f"I{group_no + 1}"]
                else:
                    successor_node = added_nodes_dict[f"end"]

                dependency_tuples_list.append((predecessor_node, current_node))
                dependency_tuples_list.append((current_node, successor_node))

        def build_parent_child_dependencies():
            config_task_names_list = []

            for task_name in aiflow_task_parent_child_dependencies.keys():
                config_task_names_list.append(task_name)
                config_task_names_list.append(aiflow_task_parent_child_dependencies[task_name][0])
                config_task_names_list.append(aiflow_task_parent_child_dependencies[task_name][1])

            config_task_names_list = [*set(config_task_names_list)]

            for task_name in config_task_names_list:
                if task_name not in dag.tasks:
                    all_nodes_list.append((DummyOperator(task_id=task_name, dag=dag), task_name))

            for config_current_task_name in aiflow_task_parent_child_dependencies.keys():
                current_node = next(node[0] for node in all_nodes_list if node[1] == config_current_task_name)
                predecessor_node = next(node[0] for node in all_nodes_list if
                                        node[1] == aiflow_task_parent_child_dependencies[config_current_task_name][0])
                successor_node = next(node[0] for node in all_nodes_list if
                                      node[1] == aiflow_task_parent_child_dependencies[config_current_task_name][1])

                dependency_tuples_list.append((predecessor_node, current_node))
                dependency_tuples_list.append((current_node, successor_node))

        # Build dependencies as per configuration
        if aiflow_task_dependencies_details:
            if "parent-child-relation" in aiflow_task_dependencies_details:
                aiflow_task_parent_child_dependencies = aiflow_task_dependencies_details["parent-child-relation"]
                build_parent_child_dependencies()

            elif "task-auto-dependencies" in aiflow_task_dependencies_details:
                task_auto_dependencies = aiflow_task_dependencies_details["task-auto-dependencies"]
                if "trigger_order" in task_auto_dependencies and \
                        task_auto_dependencies["trigger_order"].lower() == "sequential":
                    build_sequential_dependencies()
                else:
                    build_parallel_dependencies(task_auto_dependencies)
        # Build dependencies between tasks as sequential (if we don't provide any preconfigured)
        else:
            build_parallel_dependencies()

        for predecessor, successor in dependency_tuples_list:
            predecessor >> successor



import requests
import json
import csv
import io
import os
import sys
import time

# Custom Dependencies

from common_utils.scripts.creds_loader import CredentialsLoader
from common_utils.scripts.common_utilities import CommonLogging
from common_utils.scripts.s3.common_s3_utilities import CommonS3Utilities
logging = CommonLogging.get_logger()
# ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** *****


class CommonAPIUtilities:

    def __init__(self):
        pass

    @staticmethod
    def make_request(request_method, request_url, request_header, **kwargs):

        request_body_data_json = kwargs.get("request_body_data_json", None)
        request_body_data = kwargs.get("request_body_data", None)
        request_params = kwargs.get("request_params", None)
        sanitized_request_url = kwargs.get("sanitized_request_url", request_url)

        response_output = ""
        response_header = ""
        response = None
        error_details = {"error_flag": False, "error_status_code": 200, "error_msg": ""}

        try:
            logging.info(f"API request: {request_url}")
            if request_method == 'get':
                response = requests.get(
                    request_url,
                    headers=request_header, stream=True, verify=True
                )
            elif request_method == 'post':
                response = requests.post(
                    request_url,
                    json=request_body_data_json,
                    params=request_params,
                    data=request_body_data,
                    headers=request_header, stream=True, verify=True
                )

            if not response.ok:

                if (response.status_code == 500 and str(response.content).__contains__("timeout exceeded") or
                        response.status_code == 502 and str(response.content).__contains__("Bad Gateway") or
                        response.status_code == 503 and str(response.content).__contains__("Temporarily Unavailable") or
                        response.status_code == 500 and str(response.content).__contains__("Something went wrong")):
                    error_msg = "API timeout exceeded / Bad Gateway / Temporarily Unavailable : %s final_api_url %s" \
                                % (str(response.json), sanitized_request_url)
                    logging.error(error_msg)

                    error_details = {"error_flag": True,
                                     "error_status_code": response.status_code,
                                     "error_msg": error_msg}

                else:

                    error_msg = "API fetching issue (status code not ok), response_content: %s status:%s," \
                                " request_url:%s" % (str(response.content), str(response.status_code),
                                                     sanitized_request_url)
                    logging.error(error_msg)
                    raise Exception(error_msg)
            else:
                response_output = response.content
                response_header = response.headers

            return response_header, response_output, error_details
        except Exception as e:

            response_content = str(response.content) if response is not None else None
            response_header = str(response.headers) if response is not None else None
            response_status_code = str(response.status_code) if response is not None else None

            err_msg = "API fetching issue (exception), Error: %s, status:%s," \
                      " response.headers: %s," \
                      " response_content: %s" % (str(e), response_status_code, response_header, response_content)

            if str(e).lower().__contains__("InvalidChunkLength".lower()) or str(e).lower().__contains__(
                    "failed to establish".lower()) or str(e).lower().__contains__("ProxyError".lower()):
                error_details = {
                    "error_flag": True, "error_status_code": response_status_code, "error_msg": str(e)
                }

                logging.error(err_msg)
                return response_header, response_content, error_details
            else:
                logging.error(err_msg)
                raise Exception(err_msg)

    @staticmethod
    def get_api_data(request_url, request_header, **kwargs):
        request_method = kwargs.get("request_method", "get")
        request_body_data_json = kwargs.get("request_body_data_json", None)
        request_body_data = kwargs.get("request_body_data", None)
        request_params = kwargs.get("request_params", None)
        request_max_retries = kwargs.get("request_max_retries", 5)
        request_retry_delay_time = kwargs.get("request_retry_delay_time", 120)
        sanitized_request_url = kwargs.get("sanitized_request_url", request_url)  # Print URL in the Log without key
        request_response_decode_flag = kwargs.get("request_response_decode_flag", True)

        response_header, response_output, error_details = CommonAPIUtilities.make_request(
            request_method, request_url, request_header, request_body_data_json=request_body_data_json,
            request_body_data=request_body_data, request_params=request_params)
        retry_count = 1
        while error_details["error_flag"]:
            time.sleep(request_retry_delay_time)
            response_header, response_output, error_details = CommonAPIUtilities.make_request(
                request_method, request_url, request_header, request_body_data_json=request_body_data_json,
                request_body_data=request_body_data, request_params=request_params)

            if retry_count > request_max_retries:
                raise Exception(" After Retries-The server encountered an error processing the request"
                                " request_url: %s, error_status_code:%s, error_msg:%s"
                                % (sanitized_request_url, str(error_details["error_status_code"]),
                                   str(error_details["error_msg"])))
            retry_count = retry_count + 1

        if response_output is None or len(response_output) == 0:
            raise Exception("Response content is empty,request_url: %s response_text: %s"
                            % (sanitized_request_url, response_output))
        try:
            if request_response_decode_flag:
                data = response_output.decode('utf-8')
                data = data.replace("\x00", "")  # NUL char replace
                return data
            else:
                return response_output

        except Exception as e:
            partial_response_output = str(response_output)[0:100] if str(response_output) is not None else None
            err_msg = f"Data/Header to Json conversion issue, Error: {str(e)}," \
                      f"partial_response_output: {str(partial_response_output)}," \
                      f" request_url:{str(sanitized_request_url)}"
            logging.error(err_msg)
            raise Exception(err_msg)


class CommonProcessHelperUtils:

    @staticmethod
    def create_spark_session(spark_app_name, spark_options=None):
        from pyspark.sql import SparkSession
        try:
            config_options = {"spark.task.maxFailures": "1",
                                     "spark.rpc.message.maxSize": "2047"}
            if spark_options is not None:
                config_options.update(spark_options)

            if spark_app_name is None or spark_app_name == "":
                raise Exception(f"Provide Spark App Name: {str(spark_app_name)}")

            from pyspark.conf import SparkConf

            spark_conf = SparkConf().setAll([(k, v) for k, v in config_options.items()])
            spark = SparkSession.builder.appName(spark_app_name).config(conf=spark_conf).getOrCreate()
            # Set Spark configurations from the dictionary
            return spark
        except Exception as ex:
            raise Exception(f"Exception while creating spark session, Error:{str(ex)}, "
                            f"spark_options:{str(spark_options)}")

    @staticmethod
    def get_credentials(credentials_vendor_key, minio_s3_credentials):

        # *********************************************************************************
        # Reading Job Credentials
        if credentials_vendor_key is not None and credentials_vendor_key != "":
            return CredentialsLoader.get_credentials(credentials_vendor_key, minio_s3_credentials)
        else:
            return None
        # Reading Credentials
        # *********************************************************************************

    @staticmethod
    def initialize_read_arguments(spark_options=None, spark_context_log_level="INFO", mount_bucket_name="/workflows"):

        spark_app_name = sys.argv[1]
        spark = CommonProcessHelperUtils.create_spark_session(spark_app_name, spark_options)

        sc = spark.sparkContext
        sc.setLogLevel(spark_context_log_level)

        # *********************************************************************************
        # Reading Job args
        hadoop_custom_arguments = json.loads(sc.getConf().get('spark.hadoop.custom.arguments'))

        spark_job_args = hadoop_custom_arguments["spark_job_args"]
        cluster_details_spark_args = hadoop_custom_arguments["cluster_details_spark_args"]

        minio_s3_credentials = cluster_details_spark_args["s3_credentials"]
        vendor_credentials = CommonProcessHelperUtils.get_credentials(spark_job_args.get("credentials_vendor_key"),
                                                                      minio_s3_credentials)
        # Reading Job args
        # *********************************************************************************
 
        # *********************************************************************************
        # Package the project and add to the sc PyFile
        project_dir_name =  spark_job_args["project_dir_name"]
        if project_dir_name is None:
            logging.error('! project_dir_name is not provided in the spark_job_args. !!!')
            raise Exception('! project_dir_name is not provided in the spark_job_args. !!!')
        project_abspath = os.path.join(mount_bucket_name, spark_job_args["project_dir_name"])
        package_path = CommonZipHelperUtils.zip_project(project_abspath)
        logging.info(f"project_abspath: {project_abspath}")
        sc.addPyFile(package_path)
       # Package the project and add to the sc PyFile
        # *********************************************************************************

        return spark, sc, spark_app_name, spark_job_args, cluster_details_spark_args, minio_s3_credentials,\
               vendor_credentials

    @staticmethod
    def validate_source_columns_with_schema(data_record, configured_columns_details):
        data_api_cols = dict(data_record).keys()
        configured_columns_cols = dict(configured_columns_details).keys()

        no_match_cols_list = list(set(data_api_cols).symmetric_difference(set(configured_columns_cols)))

        if len(no_match_cols_list) > 0:
            err_msg = 'Columns not matching error , no_match_cols_list: %s | data_api_cols: %s ' \
                      '| columns_api_cols: %s' % (str(no_match_cols_list),
                                                  str(data_api_cols), str(configured_columns_cols))
            logging.error(err_msg)
            raise Exception(err_msg)
        else:
            logging.info("Columns matched between source data columns and pre-configured columns")

    @staticmethod
    def get_schema_from_s3(s3_schema_file_path, schema_file_object_key, s3_resource):
        data = CommonS3Utilities.read_file_from_s3(s3_resource, s3_schema_file_path,
                                                   CommonS3Utilities.MINIO_S3_PROTOCOL_PREFIX)

        if data is None or data == "":
            config_schema_json = None
        else:
            config_schema_json = json.loads(data)

        config_table_schema = config_schema_json[schema_file_object_key]
        config_table_schema = config_table_schema["schema"]

        return config_table_schema

    def get_schema_from_local(full_schema_file_path, schema_file_object_key ):
        
        local_abspath = "/" + full_schema_file_path.split("://")[1]
        try:
                with open(local_abspath, "r") as f:
                    config_schema_json = json.loads(f.read())

                    config_table_schema = config_schema_json[schema_file_object_key]
                    config_table_schema = config_table_schema["schema"]
                    return config_table_schema
        except Exception as e:
            raise Exception( "Exception occurred while reading Schema Config File: %s, Error:%s " % (local_abspath, str(e)))

    @staticmethod
    def fetch_processed_values_from_db(spark, table_field_name, output_warehouse_fq_table):
        # Fetching last ingested date from DB
        last_incr_filter_df = spark.sql(
            f"select distinct {table_field_name} as incr_filter_value from {output_warehouse_fq_table}").cache()

        db_incr_filter_value_list = last_incr_filter_df.rdd.map(lambda row: row.asDict()).collect()
        db_incr_filter_value_list = [filter_value['incr_filter_value'] for filter_value in db_incr_filter_value_list]

        logging.info("len(db_incr_filter_value_list): %s" % (str(len(db_incr_filter_value_list))))
        return db_incr_filter_value_list

    @staticmethod
    def get_latest_files_list(processed_file_names_list, input_file_paths_list):

        if len(processed_file_names_list) != 0:
            added_file_paths_list = [file_path for file_path in input_file_paths_list if
                                       os.path.basename(file_path) not in processed_file_names_list]

        else:
            logging.info("It might be first run of the job- No values in the processed list")
            added_file_paths_list = input_file_paths_list

        logging.info("Length of new / added file paths list: %s ,"
                     "" % (str(added_file_paths_list)))

        return added_file_paths_list


class CommonBasicUtilities:

    @staticmethod
    def join_prefixes_with_slash(base_uri, prefix_list):

        try:
            # Removing if any "/" character at end
            base_uri = base_uri[0:len(base_uri) - 1] if base_uri[len(base_uri) - 1] == "/" else base_uri

            api_url = base_uri
            for prefix in prefix_list:
                # Removing if any / character at start
                prefix = prefix[1:len(prefix)] if prefix[0] == "/" else prefix
                # Removing if any / character at end
                prefix = prefix[0:len(prefix) - 1] if prefix[len(prefix) - 1] == "/" else prefix
                api_url = api_url + "/" + prefix

            return api_url

        except Exception as e:
            raise Exception("Exception while joining with slashes, Error:%s, base_uri:%s,"
                            " prefix_list:%s" % (str(e), base_uri, prefix_list))

    @staticmethod
    def load_json_data(data):

        if data is None or data == "":
            raise Exception(f"Data is empty or data is None")

        try:
            response_json = json.loads(data)
            return response_json

        except Exception as e:
            partial_response_output = str(data)[0:100] if str(data) is not None else None
            err_msg = f"Data/Header to Json conversion issue, Error: {str(e)}," \
                      f"partial_response_output: {str(partial_response_output)}"
            logging.error(err_msg)
            raise Exception(err_msg)


class CommonFileHelperUtils:

    @staticmethod
    def s3_read_data_from_csv(minio_s3_credentials, s3_file_full_path, csv_delimiter):
        s3_resource = CommonS3Utilities.MinioS3Resource(minio_s3_credentials).minio_s3_resource
        try:
            logging.info("Reading the S3 file: %s" % s3_file_full_path)

            text = CommonS3Utilities.read_file_from_s3(s3_resource, s3_file_full_path,
                                                       CommonS3Utilities.MINIO_S3_PROTOCOL_PREFIX)

            csv.register_dialect('MyDialect', skipinitialspace=True, delimiter=csv_delimiter, strict=True)

            reader = csv.DictReader(io.StringIO(text), dialect='MyDialect')
            response_json = json.loads(json.dumps(list(reader)))

            return response_json

        except Exception as e:
            raise Exception("Exception occurred while reading file, Error:%s, "
                            "s3_file_full_path:%s" % (str(e), s3_file_full_path))
class CommonZipHelperUtils:
    @staticmethod
    def extract_project_name(project_abspath):
        # Remove trailing slash if present
        project_abspath = project_abspath.rstrip('/')
        project_name = os.path.basename(project_abspath)
        return project_name

    @staticmethod
    def zip_project(project_abspath):
        import zipfile
        # Zip file name
        zim_package_name = CommonZipHelperUtils.extract_project_name(project_abspath) +".zip"
        # Absolute path to where you want to save the zip file
        zip_file_abspath = os.path.join(project_abspath, zim_package_name)
        logging.info("zip_file_abspath: %s".format( str(zip_file_abspath)))
        # Create a zip file
        with zipfile.ZipFile(zip_file_abspath, 'w') as zipf:
            for root, dirs, files in os.walk(project_abspath):
                for file in files:
                    file_path = os.path.join(root, file)
                    zipf.write(file_path, os.path.relpath(file_path, project_abspath))

        return zip_file_abspath


from setuptools import setup, find_packages

# Get all directories and files recursively
def package_files(directory):
    import os
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths
 
extra_files =package_files('digixt_pipeline_template/pyspark_template')

setup(
    name='digixt-pipeline-template',
    version='1.0',
    description='DigiXT standard pipeline tempate package',
    author='Merhawi Kiflemariam',
    author_email='merhawi@saal.ai',
    packages=find_packages(),
    package_data={'': extra_files},
    entry_points={
        'console_scripts': [
            'digixt-pipeline-template=digixt_pipeline_template.__main__:main'
        ]
    },
    install_requires=[
        # Add any dependencies here
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



from datetime import datetime

# Custom Dependencies
from common_utils.scripts.common_utilities import CommonLogging
logging = CommonLogging.get_logger()
# ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** *****


class CommonDataUtilities:

    @staticmethod
    def format_array_field(record, c_name, c_info):
        target_values_array = []
        array_values_info = c_info["nested"]["$items"]
        array_values_type = array_values_info["details"]["target_type"]

        if not (record[c_name] == "" or record[c_name] is None):

            if isinstance(record[c_name], list):
                source_values_array = record[c_name]
                for value in source_values_array:
                    if array_values_type == "string":

                        if not (value == "" or value is None):
                            target_values_array.append(str(value))
                        else:
                            target_values_array.append(None)

                    elif array_values_type == "double" or array_values_type == "float":

                        if not (value == "" or value is None):
                            target_values_array.append(float(value))
                        else:
                            target_values_array.append(None)

                    elif array_values_type == "integer" or array_values_type == "long":

                        if not (value == "" or value is None):
                            target_values_array.append(int(value))
                        else:
                            target_values_array.append(None)

                    elif array_values_type == "date":
                        if not (value == "" or value is None):
                            try:
                                array_values_target_format = array_values_info['details']['target_format']
                            except KeyError as ex:
                                array_values_target_format = "%Y-%m-%d"

                            if isinstance(value, datetime):
                                formatted_value = value.date()
                            else:
                                formatted_value = datetime.strptime(value, array_values_target_format).date()
                        else:
                            formatted_value = None
                        target_values_array.append(formatted_value)

                    elif array_values_type == "boolean":

                        if value.lower() == "true":
                            target_values_array.append(True)
                        elif value.lower() == "false":
                            target_values_array.append(False)
                        else:
                            raise Exception(
                                "Boolean data value is not as expected(should be in true/false),"
                                " column_name:%s, column_value:%s"
                                % (c_name, record[c_name]))

                    elif array_values_type == "struct":

                        if value is not None:
                            if isinstance(value, dict):

                                target_values_array.append(CommonDataUtilities.format_data(value,
                                                           array_values_info["nested"]))

                            else:
                                raise Exception("Struct data value is not as expected(should be dictionary)"
                                                " column_name:%s, column_value:%s"
                                                % (c_name, record[c_name]))
                    else:
                        if not (value == "" or value is None):
                            target_values_array.append(str(value))
                        else:
                            target_values_array.append(None)

            else:
                raise Exception("Array data value is not as expected(should be List)"
                                " column_name:%s, column_value:%s"
                                % (c_name, record[c_name]))

        else:
            target_values_array = None

        return target_values_array

    @staticmethod
    def type_cast_value(map_key, map_value, map_value_type, map_info):

        if map_value_type == "string":
            format_v = str(map_value)
        elif map_value_type == "double" or map_value_type == "float":
            format_v = float(map_value)
        elif map_value_type == "integer" or map_value_type == "long":
            format_v = int(map_value)

        elif map_value_type == "boolean":
            if not (map_value == "" or map_value is None):

                if str(map_value).lower() == "true":
                    format_v = True
                elif str(map_value).lower() == "false":
                    format_v = False
                else:
                    raise Exception("Boolean data value is not as expected(should be in true/false),"
                                    " map_key:%s, column_value:%s"
                                    % (map_key, map_value))
            else:
                format_v = None
        elif map_value_type == "date":
            if not (map_value == "" or map_value is None):

                try:
                    target_format = map_info['details']['target_format']
                except KeyError as ex:
                    target_format = "%Y-%m-%d"

                if isinstance(map_value, datetime):
                    format_v = map_value.date()
                else:
                    format_v = datetime.strptime(map_value, target_format).date()
            else:
                format_v = None

        elif map_value_type == "array":
            format_v = CommonDataUtilities.format_data(map_value, map_info["nested"])
        elif map_value_type == "struct":
            format_v = CommonDataUtilities.format_data(map_value, map_info["nested"])
        else:
            raise Exception("Map value type is not in predefined list,"
                            " value:%s, map_value_type:%s" % (map_value, map_value_type))

        return format_v

    @staticmethod
    def format_map_field(record, c_name, c_info):

        map_info = c_info["nested"]["$items"]
        map_key_type = map_info["details"]["target_key_type"]
        map_value_type = map_info["details"]["target_value_type"]

        final_map = {}
        if not (record[c_name] == "" or record[c_name] is None):

            if isinstance(record[c_name], dict):

                for map_key in record[c_name].keys():

                    if map_key_type == "string":
                        formatted_key = str(map_key)
                    else:
                        raise Exception("Map key type should be string, c_name:%s, map_key_type:%s"
                                        % (c_name, map_key_type))

                    final_map[formatted_key] = CommonDataUtilities.type_cast_value(
                        map_key, record[c_name][map_key], map_value_type, map_info)

            else:
                raise Exception("Map field is not of type dictionary,"
                                "c_name:%s, record[c_name]:%s" % (str(c_name), str(record[c_name])))

        else:
            final_map = None

        return final_map

    @staticmethod
    def format_data(record, table_schema):

        formatted_record = {}
        for c_name, c_info in table_schema.items():

            target_name = c_info['details']['target_name']
            target_type = c_info['details']['target_type']
            # formatted_record[target_name] = record[c_name]

            try:

                if target_type == "string":
                    formatted_record[target_name] = str(record[c_name]) \
                        if not (record[c_name] == "" or record[c_name] is None) else None

                elif target_type == "double" or target_type == "float":

                    formatted_record[target_name] = float(record[c_name]) \
                        if not (record[c_name] == "" or record[c_name] is None) else None

                elif target_type == "integer" or target_type == "long":
                    formatted_record[target_name] = int(record[c_name]) \
                        if not (record[c_name] == "" or record[c_name] is None) else None

                elif target_type == "date":
                    if not (record[c_name] == "" or record[c_name] is None):

                        try:
                            target_format = c_info['details']['target_format']
                        except KeyError as ex:
                            target_format = "%Y-%m-%d"

                        if isinstance(record[c_name], datetime):
                            formatted_record[target_name] = record[c_name].date()
                        else:
                            formatted_record[target_name] = datetime.strptime(record[c_name], target_format).date()

                    else:
                        formatted_record[target_name] = None

                elif target_type == "timestamp":
                    if not (record[c_name] == "" or record[c_name] is None):

                        try:
                            target_format = c_info['details']['target_format']
                        except KeyError as ex:
                            target_format = "%Y-%m-%d %H:%M:%S"

                        if isinstance(record[c_name], datetime):
                            formatted_record[target_name] = record[c_name]
                        else:
                            formatted_record[target_name] = datetime.strptime(record[c_name], target_format)

                    else:
                        formatted_record[target_name] = None

                elif target_type == "boolean":

                    if not (record[c_name] == "" or record[c_name] is None):

                        if str(record[c_name]).lower() == "true":
                            formatted_record[target_name] = True
                        elif str(record[c_name]).lower() == "false":
                            formatted_record[target_name] = False
                        else:
                            raise Exception("Boolean data value is not as expected(should be in true/false),"
                                            " column_name:%s, column_value:%s"
                                            % (c_name, record[c_name]))
                    else:
                        formatted_record[target_name] = None

                elif target_type == "struct":
                    if record[c_name] is not None:
                        if isinstance(record[c_name], dict):

                            formatted_record[target_name] = CommonDataUtilities.format_data(
                                record[c_name], c_info["nested"])

                        else:
                            raise Exception("Struct data value is not as expected(should be dictionary)"
                                            " column_name:%s, column_value:%s"
                                            % (c_name, record[c_name]))

                elif target_type == "array":

                    formatted_record[target_name] = CommonDataUtilities.format_array_field(record, c_name, c_info)

                elif target_type == "map":
                    formatted_record[target_name] = CommonDataUtilities.format_map_field(record, c_name, c_info)

                else:
                    err_msg = 'Data type mapping error, data type is not available in predefined list' \
                              'record: %s, column_name:%s, column_type %s ' % (
                                  str(record), str(c_name), str(target_type))
                    logging.error(err_msg)
                    raise Exception(err_msg)

            except Exception as e:
                err_msg = "Type conversion/ key issue,  error %s, record_column_name: %s, record_column_value: %s" \
                          % (str(e), str(c_name), record[c_name])
                logging.error(err_msg)
                raise Exception(err_msg)

        return formatted_record

    @staticmethod
    def generate_spark_array_schema(c_info):
        field = {"containsNull": True, "type": "array"}

        array_values_info = c_info["nested"]["$items"]
        array_values_type = array_values_info["details"]["target_type"]

        if array_values_type in ("string", "double", "float", "integer", "long", "boolean", "date", "timestamp"):
            field['elementType'] = array_values_type
        elif array_values_type == "struct":
            field["elementType"] = CommonDataUtilities.create_spark_schema(array_values_info["nested"])

        return field

    @staticmethod
    def generate_spark_map_field_schema(c_info):
        sub_field = {"valueContainsNull": True, "type": "map"}

        map_info = c_info["nested"]["$items"]
        map_key_type = map_info["details"]["target_key_type"]
        map_value_type = map_info["details"]["target_value_type"]

        if map_key_type == "string":
            sub_field['keyType'] = "string"
        else:
            raise Exception("Map key type is not string type, map_key_type:%s" % map_key_type)

        if map_value_type in ("string", "double", "float", "integer", "long", "boolean", "date"):
            sub_field['valueType'] = map_value_type
        elif map_value_type == "struct":
            sub_field['valueType'] = CommonDataUtilities.create_spark_schema(map_info["nested"])
        else:
            raise Exception("Map value type is not in the predefined list, map_key_type:%s" % map_value_type)

        return sub_field

    @staticmethod
    def create_spark_schema(table_schema):
        fields = []

        for c_name, c_info in table_schema.items():
            field = {}
            target_name = c_info['details']['target_name']
            target_type = c_info['details']['target_type']

            field['name'] = target_name
            field['nullable'] = True
            field['metadata'] = {}
            try:

                if target_type in ("string", "double", "float", "integer", "long", "boolean", "date", "timestamp"):
                    field['type'] = target_type
                elif target_type == "struct":

                    field['type'] = CommonDataUtilities.create_spark_schema(c_info["nested"])

                elif target_type == "array":
                    field['type'] = CommonDataUtilities.generate_spark_array_schema(c_info)
                    pass

                elif target_type == "map":
                    field['type'] = CommonDataUtilities.generate_spark_map_field_schema(c_info)

                else:
                    err_msg = 'Data type  error, data type is not available in predefined list' \
                              'column_nam: %s, column_type %s ' % (str(c_name), str(target_type))
                    logging.error(err_msg)
                    raise Exception(err_msg)
                fields.append(field)
            except Exception as e:
                err_msg = "Spark schema creation issue,  error %s, record_column_name: %s" % (str(e), str(c_name))
                logging.error(err_msg)
                raise Exception(err_msg)

        spark_schema = {"fields": fields, "type": "struct"}
        return spark_schema



import logging
import yaml
import boto3
from botocore.config import Config
from botocore.exceptions import ClientError

from yaml.loader import SafeLoader

logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%y/%m/%d %H:%M:%S')

logging = logging.getLogger()

creds_bucket = 'system'
creds_path = '/config/cr.yaml'


class CredentialsLoader:
    creds_file = None

    def __init__(self):
        pass
    
    ## S3
    @staticmethod
    def get_s3_client(s3_credentials):
        s3_host = s3_credentials['s3_host']
        s3_access_key = s3_credentials['s3_access_key']
        s3_secret_access_key = s3_credentials['s3_secret_access_key']

        s3 = boto3.resource(
            's3',
            endpoint_url=s3_host,
            aws_access_key_id=s3_access_key,
            aws_secret_access_key=s3_secret_access_key,
            config=Config(signature_version='s3v4'),
            region_name='us-east-1'
        )
        return s3
        
    @staticmethod
    def get_s3_file(bucket, object_name, s3_credentials):
        try:
            s3 = CredentialsLoader.get_s3_client(s3_credentials)
            obj = s3.Bucket(bucket).Object(object_name)
            data = obj.get()['Body'].read().decode('utf-8')
            return data
        except ClientError as e:
            print(e)
        return None

    @staticmethod
    def get_credentials(vendor_name, s3_credentials=None, mounted_path=None):
        if CredentialsLoader.creds_file is None:
            if s3_credentials:
                CredentialsLoader.creds_file = CredentialsLoader.get_s3_file(creds_bucket, creds_path, s3_credentials)
            else:
                with open(mounted_path, 'r') as file:
                    CredentialsLoader.creds_file = file.read()

        try:
            credentials = yaml.load(CredentialsLoader.creds_file, Loader=SafeLoader)
            return credentials[vendor_name]

        except yaml.scanner.ScannerError as sc:
            print(f'!!! YAML is not readable. Problem clue: {sc.problem}')
            print(f'!!! For more troubleshooting, take the file and try to load it manually in localhost.')
            raise Exception('!!! Impossible to proceed without credentials!!!')
        except Exception:
            print('!!! YAML is not readable. Please check if it is well formatted!')
            raise Exception('!!! Impossible to proceed without credentials!!!')
